/*************************************************************************************
    Semantic check about HPF distribute and align directive
    written by Wang Zhenlin
**************************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include <stdio.h>

#   include "typedef.h"
#   include "bif.h"
#   include "ll.h"
#   include "symb.h"

#include "macro.h"
#include "defnodes.h"
#include "tags.h"
#include "decl.h"
 
#include "exglobal.h"

/* check processors statement */
int processors_decl_stmtc(cur_bfnd)
PTR_BFND cur_bfnd;
{
	PTR_LLND expr_list,ele,range;
	PTR_SYMB sym;

	for (expr_list=cur_bfnd->entry.Template.ll_ptr1;expr_list;
		expr_list=expr_list->entry.Template.ll_ptr2 )
	{
		ele=expr_list->entry.Template.ll_ptr1;
		sym = NODE_SYMB(ele);
		range=NODE_OPERAND0(ele);
		  /* check range if present in dimension */
                check_array_range(range);
		sym->decl=HARD;
	}
}

/*check distribute statement*/
int distribute_decl_smtc(cur_bfnd)
PTR_BFND cur_bfnd;
{
   distribute_attr_smtc(cur_bfnd->entry.Template.ll_ptr1,cur_bfnd->entry.Template.ll_ptr3->entry.Template.ll_ptr1,1); 
/*see distribute.c*/
}

/*check redistribute statement*/
int redistribute_stmt_smtc(cur_bfnd)
PTR_BFND cur_bfnd;
{
   distribute_attr_smtc(cur_bfnd->entry.Template.ll_ptr1,cur_bfnd->entry.Template.ll_ptr3->entry.Template.ll_ptr1,0); /*see distribute.c*/
}

/*check align statement*/
int align_decl_smtc(cur_bfnd)
PTR_BFND cur_bfnd;
{
				/*    1                  2: ALIGN_OP */
   align_attr_smtc(cur_bfnd->entry.Template.ll_ptr1,cur_bfnd->entry.Template.ll_ptr3->entry.Template.ll_ptr1,1); /*see align.c*/
}

/*check range statement*/
int range_decl_smtc(cur_bfnd)
PTR_BFND cur_bfnd;
{
				/*    1                  2: ALIGN_OP */
   range_attr_smtc(cur_bfnd->entry.Template.ll_ptr1,cur_bfnd->entry.Template.ll_ptr3->entry.Template.ll_ptr1,1); 
}

/*check range statement*/
int shadow_decl_smtc(cur_bfnd)
PTR_BFND cur_bfnd;
{
				/*    1                  2: ALIGN_OP */
   shadow_attr_smtc(cur_bfnd->entry.Template.ll_ptr1,cur_bfnd->entry.Template.ll_ptr3->entry.Template.ll_ptr1,1); 
}

/*check realign statement*/
int realign_stmt_smtc(cur_bfnd)
PTR_BFND cur_bfnd;
{
   align_attr_smtc(cur_bfnd->entry.Template.ll_ptr1,cur_bfnd->entry.Template.ll_ptr3->entry.Template.ll_ptr1,0); /*see align.c*/
}

/*check combined_directive statement*/
int combined_directive_decl_smtc(cur_bfnd)
PTR_BFND cur_bfnd;
{
   PTR_LLND  attr_list,attr,var_list,var_ele,range;
   PTR_SYMB  sym;
   PTR_TYPE  type;
   int	     attrbit=0;
   
   var_list = cur_bfnd->entry.Template.ll_ptr1;
					/* attr_list ->  EXPR_LIST */
   attr_list = cur_bfnd->entry.Template.ll_ptr3;
   while (attr_list != LLNULL){
      attr = attr_list->entry.Template.ll_ptr1;
      switch (attr->variant) {
					/* modified by zheng.gb & HIT */
          case DISTRIBUTE_ATT:
		  if (attrbit & 0x01 )  error1(327);
		  attrbit |= 0x01;
                  distribute_attr_smtc(cur_bfnd->entry.Template.ll_ptr1,
					attr->entry.Template.ll_ptr1,1);       
                  break; 
					/* modified by zheng.gb & HIT */
          case ALIGN_ATT:
				/* by zheng.gb ???? miss a argument */
		  if (attrbit & 0x02 )  error1(327);
		  attrbit |= 0x02;
                  align_attr_smtc(cur_bfnd->entry.Template.ll_ptr1,
					attr->entry.Template.ll_ptr1,1);       
                  break;
          case INHERIT_OP:
		  if (attrbit & 0x04 )  error1(327);
		  attrbit |= 0x04;
                  inherit_attr_smtc(cur_bfnd->entry.Template.ll_ptr1);
                  break;
          case DYNAMIC_OP:
   		  /* added by LY, check for PURE function */
		  if (attrbit & 0x08 )  error1(327);
		  attrbit |= 0x08;
   		  dynamic_attr_smtc(cur_bfnd->entry.Template.ll_ptr1);
                  break;
	  case  DIMENSION_OP:
		  if (attrbit & 0x10 ) error1(327);
		  attrbit |= 0x10 ;
		  break;
	  case  TEMPLATE_VAL:
		  if (attrbit & 0x20 ) error1(327);
		  attrbit |= 0x20 ;
		  break;
	  case  PROCESSORS_VAL:
		  if (attrbit & 0x40 ) error1(327);
		  attrbit |= 0x40 ;
		  break;
	/* for HPF 2.0 */
	  case ACTIVE_OP:
		  if (attrbit & 0x80 ) error1(327);
		  attrbit |= 0x80 ;
		  break;
	  case SHADOW_OP:
		  if (attrbit & 0x100 ) error1(327);
		  attrbit |= 0x100 ;
                  shadow_attr_smtc(cur_bfnd->entry.Template.ll_ptr1,
					attr->entry.Template.ll_ptr1,1);       
		  break;
	  case DIR_RANGE_ATTR:
		  if (attrbit & 0x200 ) error1(327);
		  attrbit |= 0x200 ;
                  range_attr_smtc(cur_bfnd->entry.Template.ll_ptr1,
					attr->entry.Template.ll_ptr1,1);       
		  break;
	  default:
		  printf("Unkown combine diective attr :%d!\n",attr->variant);
		  break;
      }
      attr_list = NODE_OPERAND1(attr_list);
   }

   for ( ; var_list ; var_list=NODE_OPERAND1(var_list)) {
	var_ele=var_list->entry.Template.ll_ptr1;
	sym=get_var_name(var_ele);
	if (sym) {
		sym->decl=HARD;
			/* check if dimension is specified */
		if (attrbit & 0x10 )
		{
			type=get_base_type(sym->type);
			if ( type->variant != T_TEMPLATE && type->variant !=T_PROCESSORS )
				error3(328,sym->ident);
		}
	}
	if ( var_ele->variant == ARRAY_REF)
	{
		range=var_ele->entry.Template.ll_ptr1;
		/* check range if present in dimension */
                check_array_range(range);
	}
   }
}

int inherit_decl_smtc(cur_bfnd)
PTR_BFND cur_bfnd;
{
   inherit_attr_smtc(cur_bfnd->entry.Template.ll_ptr1);
}

int dynamic_decl_smtc(cur_bfnd)
PTR_BFND cur_bfnd;
{
   dynamic_attr_smtc(cur_bfnd->entry.Template.ll_ptr1);
}

int template_stmt(cur_bfnd)
PTR_BFND cur_bfnd;
{
	PTR_LLND expr_list ,expr;

	for (expr_list = BIF_LL1(cur_bfnd) ; expr_list ; expr_list = NODE_OPERAND1(expr_list))
	{
		expr=NODE_OPERAND0(expr_list);
		NODE_SYMB(expr)->decl=HARD;
	}
}

/*variable which has INHERIT attribute should be dummy argument*/
int inherit_attr_smtc(var_list)
PTR_LLND var_list;
{
   PTR_SYMB var;

   while (var_list != LLNULL) {
					/* add by zheng.gb */
       check_leaf_sym(var_list);
       var = var_list->entry.Template.ll_ptr1->entry.Template.symbol;
       var->attr |= INHERIT_BIT;
       if (!is_dummy(cur_scope(),var))
           error3(308,var->ident);
       if (have_attribute(var,ALIGN_BIT))
           error3(309,var->ident);
       var_list = var_list->entry.Template.ll_ptr2;
   } 
   /* added by LY, check for PURE function */
   if (is_in_pure())
	error3(320);
}

/*variable which has DYNAMIC attribute should be dummy argument*/
int dynamic_attr_smtc(var_list)
PTR_LLND var_list;
{
   PTR_SYMB var;

   while (var_list != LLNULL) {
					/* add by zheng.gb */
       check_leaf_sym(var_list->entry.Template.ll_ptr1);
       var = var_list->entry.Template.ll_ptr1->entry.Template.symbol;
       var->attr |= DYNAMIC_BIT;
       var_list = var_list->entry.Template.ll_ptr2;
   } 
   /* added by LY, check for PURE function */
   if (is_in_pure())
	error3(321);
}

int isSubobject(sym)
PTR_LLND sym;
{
   PTR_SYMB symbol=sym->entry.Template.symbol;
   PTR_TYPE basetype;

   basetype=get_base_type(symbol->type);
   if (basetype->variant == T_TEMPLATE ) return 0;
   if (sym->variant == VAR_REF) return 0;
   return 1;
}

int have_attribute(symbol,attr) 
PTR_SYMB symbol;
int attr;
{
   if (symbol->attr & attr)
      return 1;
   else
      return 0;
}

/*
int independent_stmt(cur_bfnd)
PTR_BFND cur_bfnd;
{
	PTR_LLND  new_list = BIF_LL1(cur_bfnd);
	PTR_LLND  reduce_list = BIF_LL2(cur_bfnd);
	PTR_LLND  expr_list,exp;

	for (expr_list = reduce_list ; expr_list ; expr_list=NODE_OPERAND1(expr_list) )
	{
		PTR_SYMB reduce_sym = NODE_SYMB(NODE_OPERAND0(expr_list));
		if (reduce_sym->entry.var_decl.hpfattr)
			error3(650,reduce_sym->ident);
		for (exp = new_list ; exp ; exp=NODE_OPERAND1(exp) )
		{
			PTR_SYMB new_sym=NODE_SYMB(NODE_OPERAND0(exp));
			if (new_sym == reduce_sym)
				error3(651,reduce_sym->ident);
		}
	}	
}
*/

int range_attr_smtc(var_list,range_attr)
PTR_LLND var_list;
PTR_LLND range_attr;
{
	PTR_LLND range_list=range_attr;
	int 	 olddim = -1,dim;
	int 	num = 0;
	
	for ( ; var_list ; var_list=NODE_OPERAND1(var_list) )
	{
		PTR_LLND var = NODE_OPERAND0(var_list);
		PTR_SYMB sym = NODE_SYMB(var);
		/* get the dimension of a symbol */
		dim = get_dimension(sym);
		if (olddim == -1 )
			olddim=dim;
		else if (olddim != dim) {
			error1(658); return;
		}
			/* check if alignee */
		if (sym->attr & ALIGN_BIT ) error3(661,sym->ident);
			/* check symbol attributs */
		if (sym->attr & DYNAMIC_BIT ) continue;
		if (sym->attr & INHERIT_BIT)
			if ((sym->attr & DISTRIBUTE_BIT) == 0) continue;
		if (sym->entry.var_decl.hpfattr)
		{
			PTR_LLND op=sym->entry.var_decl.hpfattr;
			if (op->variant == DISTRIBUTE_OP)
				if (NODE_OPERAND0(NODE_OPERAND0(op))) continue;
		}
		/* else report error */
		error3(660 , sym->ident);
	
	}
	/*  traverse along the range list */
	for ( ; range_list ; range_list=NODE_OPERAND1(range_list) )
	{
		PTR_LLND range_attr=NODE_OPERAND0(range_list) ;
		PTR_LLND range_format_list=NODE_OPERAND0(range_attr) ;
		int count =0;
		num++;
		for ( ; range_format_list ; range_format_list=NODE_OPERAND1(range_format_list) )
			count++;
		if (count != dim) error2(659,num);
	}
}


int shadow_attr_smtc(var_list,shadow_attr)
PTR_LLND var_list;
PTR_LLND shadow_attr;
{
	PTR_LLND shadow_list=shadow_attr;
	
	for ( ; var_list ; var_list=NODE_OPERAND1(var_list) )
	{
	}
	/*  traverse along the range list */
	for ( ; shadow_list ; shadow_list=NODE_OPERAND1(shadow_list) )
	{
		PTR_LLND ddot=NODE_OPERAND0(shadow_list) ;
		PTR_LLND rangeop = NODE_OPERAND0(ddot);
		PTR_LLND low_width = NODE_OPERAND0(rangeop);
		PTR_LLND high_width = NODE_OPERAND1(rangeop);
		if (low_width)
		{
			if ( check_spec_expr(low_width) != 0 )
				error1(662);
			else if (check_initialization_expr(low_width))
				{
					comput_const_expr(low_width,0);
					if (low_width->const_val < 0) error1(662);
				}
		}
		if (high_width)
		{
			if ( check_spec_expr(high_width) != 0 )
				error1(662);
			else if (check_initialization_expr(high_width))
				{
					comput_const_expr(high_width,0);
					if (high_width->const_val < 0) error1(662);
				}
		}
	}
}


/*  HPF extentions by our group */
/* Zheng Gengbin */

int outofcore_decl_smtc(cur_bfnd)
PTR_BFND cur_bfnd;
{
	PTR_LLND expr_list,ele,range;
	PTR_SYMB sym;

	for (expr_list=BIF_LL1(cur_bfnd); expr_list; expr_list=NODE_OPERAND1(expr_list) )
	{
		ele=NODE_OPERAND0(expr_list);
		sym=ele->entry.Template.symbol;
		if (is_scalar(sym->type)) {
		   error3(854, sym->ident);
		}
		if (sym->attr & OUTOFCORE_BIT) {
		   error3(860, sym->ident);
		}
		sym->attr |= OUTOFCORE_BIT;
	}
}

int memory_decl_smtc(cur_bfnd)
PTR_BFND cur_bfnd;
{
   PTR_LLND mem_size = BIF_LL1(cur_bfnd);
   int status;
   static int val = -1;

     /* must declared in main program */
   if (BIF_CODE(BIF_CP(cur_bfnd)) != PROG_HEDR)
   {
	error1(855);
	return 0;
   }

   if (val != -1) {
	error1(853);
	return 0;
   }
   check_expr_type(mem_size);	
   if (isInteger(mem_size) && is_scalar(mem_size->type)) {
	status = comput_const_expr(mem_size,1);
	if (status!=0) {
	    error1(850);
	    return 0;
	};
 	if (mem_size->const_val < 0) {
	    error1(851);
	    return 0;
	}
   }           /* if */
   val = mem_size->const_val;
   return 1;
}


int incoresize_decl_smtc(cur_bfnd)
PTR_BFND cur_bfnd;
{
   PTR_LLND size = BIF_LL1(cur_bfnd);
   int status;
   static int val = -1;

   if (BIF_CODE(BIF_CP(cur_bfnd)) != PROG_HEDR)
   {
	error1(856);
	return 0;
   }

   if (val != -1) {
	error1(852);
	return 0;
   }
   check_expr_type(size);	
   if (isInteger(size) && is_scalar(size->type)) {
	status = comput_const_expr(size,1);
	if (status!=0) {
	    error1(851);
	    return 0;
	};
 	if (size->const_val < 0) {
	    error1(851);
	    return 0;
	}
   }           /* if */
   val = size->const_val;
   return 1;
}



int major_decl_smtc(cur_bfnd)
PTR_BFND cur_bfnd;
{
	PTR_LLND expr_list,ele,range;
	PTR_SYMB sym;

	for (expr_list=BIF_LL1(cur_bfnd); expr_list; expr_list=NODE_OPERAND1(expr_list) )
	{
		ele=NODE_OPERAND0(expr_list);
		check_leaf_sym(ele);
		sym=ele->entry.Template.symbol;
		if (is_scalar(sym->type)) {
		   error3(857, sym->ident);
		}
		if ( (sym->attr & OUTOFCORE_BIT)==0) {
	   		error3(861, sym->ident);
		}
	}
}


int on_stmt_smtc(cur_bfnd)
PTR_BFND cur_bfnd;
{
 	PTR_LLND exprlist,var,fileexpr,homeop,residentop;
        PTR_SYMB sym;
	PTR_TYPE type;

	 /* check ON clause */
	homeop = BIF_LL1(cur_bfnd);
	var = NODE_OPERAND0(homeop);
	check_leaf_sym(var);
	sym = NODE_SYMB(var);
	type = sym->type;
	if (type->variant == T_ARRAY) 
		type=type->entry.ar_decl.base_type;
	
	switch (homeop->variant) {
	case HOME_PROCS:
	   if (type->variant != T_PROCESSORS && type->variant != T_TEMPLATE)
		error3(862, SYMB_IDENT(sym));
	   break;
	case HOME_OP:
	   if (type->variant == T_PROCESSORS || type->variant == T_TEMPLATE)
		error3(863, SYMB_IDENT(sym));
	   break;
	}

	 /* check RESIDENT clause */
	residentop = BIF_LL2(cur_bfnd);
	if (residentop)
	{
	exprlist = NODE_OPERAND0(residentop);
   	for ( ; exprlist; exprlist=NODE_OPERAND1(exprlist) ) {
	   var = NODE_OPERAND0(exprlist);
	   check_leaf_sym(var);
	}
	}
}

int on_block_stmt_smtc()
{
        PTR_BLOB  blob;
	on_stmt_smtc(cur_bfnd);
	blob = cur_bfnd->entry.Template.bl_ptr1;
	check_a_block(blob);
	cur_bfnd=cur_bfnd->thread;
}

int resident_stmt_smtc(cur_bfnd)
PTR_BFND cur_bfnd;
{
 	PTR_LLND exprlist,var,fileexpr,homeop,residentop;

	 /* check RESIDENT clause */
	residentop = BIF_LL1(cur_bfnd);
	exprlist = NODE_OPERAND0(residentop);
   	for ( ; exprlist; exprlist=NODE_OPERAND1(exprlist) ) {
	   var = NODE_OPERAND0(exprlist);
	   check_leaf_sym(var);
	}
}

int resident_block_stmt_smtc()
{
        PTR_BLOB  blob;
	resident_stmt_smtc(cur_bfnd);
	blob = cur_bfnd->entry.Template.bl_ptr1;
	check_a_block(blob);
	cur_bfnd=cur_bfnd->thread;
}

int task_region_stmt_smtc()
{
        PTR_BLOB  blob;
	blob = cur_bfnd->entry.Template.bl_ptr1;
	check_a_block(blob);
	cur_bfnd=cur_bfnd->thread;
}

int filemap_stmt_smtc(cur_bfnd)
PTR_BFND cur_bfnd;
{
        PTR_LLND var,fileexpr,op,major_op;
        PTR_SYMB sym;

	op = BIF_LL1(cur_bfnd);
	var =  NODE_OPERAND0(op);
	fileexpr = NODE_OPERAND1(op);
	major_op = BIF_LL2(cur_bfnd);

	check_leaf_sym(var);
	sym = NODE_SYMB(var);
	if ( op->variant == EQ_OP && (sym->attr & OUTOFCORE_BIT)==0) {
	   error3(861, sym->ident);
	}
	if (is_scalar(sym->type)) {
	   error3(858, sym->ident);
	}
	check_expr_type(fileexpr);
	if (isString(fileexpr) == 0) {
	   error1(859);
	}
	if (major_op) {
	   check_expr_type(major_op);
	   if (op->variant == EQ_OP) {
	      error1(865);
	   }	
	   if ((major_op->attr & PARAMETER_BIT) == 0) {
	      error1(865);
	   }
	   else {
	      if (strcasecmp(major_op->entry.string_val, "\"row\"") &&
		  strcasecmp(major_op->entry.string_val, "\"column\""))
	      error1(865);
	   }
	}
}


int io_processor_stmt(cur_bfnd)
PTR_BFND cur_bfnd;
{
        PTR_LLND exprlist,var,range;
        PTR_SYMB sym;

	exprlist = BIF_LL1(cur_bfnd);
   	for ( ; exprlist ; exprlist=NODE_OPERAND1(exprlist) )
	{
	   var =  NODE_OPERAND0(exprlist);
	   check_leaf_sym(var);
/*
	   range=NODE_OPERAND0(var);
	   check_array_range();
*/
	   sym = NODE_SYMB(var);
	   if (var->type->variant != T_PROCESSORS)
		error3(864,sym->ident);
	}
}



